// WARNING: this sample is under construction! Use it on your own risk.
#if defined _MSC_VER && _MSC_VER >= 1400
#pragma warning(disable : 4100)
#endif


#include <iostream>
#include <iomanip>
#include <assert.h>

// #include "opencv2/contrib/contrib.hpp"
#include "opencv2/objdetect/objdetect.hpp"
#include "opencv2/highgui/highgui.hpp"
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/gpu/gpu.hpp"

using namespace std;
using namespace cv;
using namespace cv::gpu;


void help()
{
    cout << "Usage: ./face_detect " << endl; 
    cout << "\t--cascade <cascade_file> " << endl; 
    cout << "\t(<image>|--video <video>)" << endl;
    cout << "\t--output <output_file> " << endl; 
    cout << "Using OpenCV version " << CV_VERSION << endl << endl;
}


template<class T>
void convertAndResize(const T& src, T& gray, T& resized, double scale)
{
    if (src.channels() == 3)
    {
        cvtColor( src, gray, CV_BGR2GRAY );
    }
    else
    {
        gray = src;
    }

    Size sz(cvRound(gray.cols * scale), cvRound(gray.rows * scale));

    if (scale != 1)
    {
        resize(gray, resized, sz);
    }
    else
    {
        resized = gray;
    }
}


void matPrint(Mat &img, int lineOffsY, Scalar fontColor, const string &ss)
{
    int fontFace = FONT_HERSHEY_DUPLEX;
    double fontScale = 0.8;
    int fontThickness = 2;
    Size fontSize = cv::getTextSize("T[]", fontFace, fontScale, fontThickness, 0);

    Point org;
    org.x = 1;
    org.y = 3 * fontSize.height * (lineOffsY + 1) / 2;
    putText(img, ss, org, fontFace, fontScale, CV_RGB(0,0,0), 5*fontThickness/2, 16);
    putText(img, ss, org, fontFace, fontScale, fontColor, fontThickness, 16);
}

int main(int argc, const char *argv[])
{
    // Check that we have enough arguments.  
    if (argc == 1) { help(); return -1;}
    // Check that there is a CUDA device.  
    assert(getCudaEnabledDeviceCount() > 0 && "No GPU found or the library is compiled without GPU support");
    // Print the information about this CUDA device.  
    cv::gpu::printShortCudaDeviceInfo(cv::gpu::getDevice());

    string cascadeName;
    string inputName;

    // Flags to determine whether this is an input image or video.  
    bool isInputImage = false;
    bool isInputVideo = false;

    // Parse the arguments. 
    for (int i = 1; i < argc; ++i)
    {
        if (string(argv[i]) == "--cascade")
            cascadeName = argv[++i];
        else if (string(argv[i]) == "--video")
        {
            inputName = argv[++i];
            isInputVideo = true;
        }
        else if (string(argv[i]) == "--help")
        {
            help();
            return -1;
        }
        else if (!isInputImage)
        {
            inputName = argv[i];
            isInputImage = true;
        }
        else
        {
            cout << "Unknown key: " << argv[i] << endl;
            return -1;
        }
    }

    CascadeClassifier_GPU cascade_gpu;
    assert(cascade_gpu.load(cascadeName) && "ERROR: Could not load cascade classifier. "); 


    VideoCapture capture;
    Mat image;

    if (isInputImage)
    {
        // Read an image.  
	cout << "Found Image File: " << endl; 
	image = imread(inputName);
        CV_Assert(!image.empty());
    } else {
        // Read the video.  
	cout << "Found Video File: " << endl; 
        capture.open(inputName);
        CV_Assert(capture.isOpened());
    }

    Mat frame, frame_cpu, faces_downloaded, frameDisp;
    GpuMat frame_gpu, gray_gpu, resized_gpu, facesBuf_gpu;

    /* parameters */
    double scaleFactor = 1.0;
    bool findLargestObject = false;
    bool filterRects = true;
    bool helpScreen = false;

    int detections_num;
    for (;;)
    {
        // If we have marked this as a video than we need to retrieve the frame. 
        if (isInputVideo)
        {
            capture >> frame;
            if (frame.empty()) break;
        }
    
        frame_gpu.upload(image.empty() ? frame : image);
        
        // Convert to gray scale and scale the image.  
        convertAndResize(frame_gpu, gray_gpu, resized_gpu, scaleFactor);
       
        cascade_gpu.visualizeInPlace = true;
        cascade_gpu.findLargestObject = findLargestObject;
        detections_num = cascade_gpu.detectMultiScale(resized_gpu, facesBuf_gpu, 1.2, (filterRects || findLargestObject) ? 4 : 0);
        facesBuf_gpu.colRange(0, detections_num).download(faces_downloaded);

        //print detections to console
        if ((filterRects || findLargestObject) && detections_num > 0)
        {
            Rect *faceRects = faces_downloaded.ptr<Rect>();
            for (int i = 0; i < min(detections_num, 2); ++i)
            {
                cout << ", [" << setw(4) << faceRects[i].x
                     << ", " << setw(4) << faceRects[i].y
                     << ", " << setw(4) << faceRects[i].width
                     << ", " << setw(4) << faceRects[i].height << "]";
            }
        }
        cout << endl;

	    if (isInputImage) {
	    	break;
	    }

    }

    return 0;
}
